@@ -432,14 +432,14 @@ unsigned int run_ntlm_auth(const char *username,
 	}
 
         /* parent */
-        if (close(child_out[0]) == -1) {
-                close(child_in[1]);
+        if (fclose(pipe_out) == -1) {
+                fclose(pipe_in);
                 notice("error closing pipe?!? for child OUT[0]");
                 return NOT_AUTHENTICATED;
         }
 
        /* parent */
-        if (close(child_in[1]) == -1) {
+        if (fclose(pipe_in) == -1) {
                 notice("error closing pipe?!? for child IN[1]");
                 return NOT_AUTHENTICATED;
         }
